@@ -1,12 +1,12 @@
 /* darkstat 3
- * copyright (c) 2006, 2007 Emil Mikulic.
+ * copyright (c) 2006-2011 Emil Mikulic.
  *
  * graph_db.h: round robin database for graph data
  */
+#ifndef __DARKSTAT_GRAPH_DB_H
+#define __DARKSTAT_GRAPH_DB_H
 
-#ifdef linux
-# include <stdint.h> /* for uint64_t */
-#endif
+#include <stdint.h> /* for uint64_t on Linux and OS X */
 
 enum graph_dir {
    MIN_GRAPH_DIR = 1,
@@ -26,4 +26,5 @@ int graph_export(const int fd);
 struct str *html_front_page(void);
 struct str *xml_graphs(void);
 
+#endif
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
